import struct
import hashlib

class Error(Exception): pass
class BadSID(Error): pass

SIDHEADERv1 = ">HHHHHHHI32s32s32s"
SIDHEADERv2 = ">HBBH"

driver_code = [
	# Cold start (at 0xff00)
	0xa2, 0xFF,			# 00: LDX $#FF, set stack pointer
	0x78,				# 02: SEI, disable interrupts (already off after reset)
	0x9a,				# 03: TXS, write to stack pointer
	0xd8,				# 04: CLD, disable decimal mode
	0xa9, 0x3,			# 05: LDA x, song #x in A (x = 3 here)
	0x20, 0x00, 0x7f, 	# 07: JSR 0x7f00, call song init function
	0xea,				# 0a: NOP, turn this into BRK for debugging
	0x58,				# 0b: CLI, enable interrupts
	0x4c, 0x0c, 0xff,   # 0c: JMP 0xff0c, endless loop

	# IRQ handler (at 0xff0f)
	0x20, 0x03, 0x7f,	# 0f: JSR play routine (0x7f03 here)
	0x40,				# 12: RTI
	# Unused, just here for padding.
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
	# Vectors:
	0x0, 0xff, # Cold start (reset) = 0xff00
	0xf, 0xff  # IRQ = 0xff0f
]

def makeDriver(songInitAddr, songIRQAddr, songNumber, brkAfterInit = False):
	driver = driver_code[:]
	assert len(driver) == 256, len(driver)
	driver[6] = songNumber
	driver[8] = songInitAddr & 0xff
	driver[9] = (songInitAddr & 0xff00) >> 8
	driver[0x10] = songIRQAddr & 0xff
	driver[0x11] = (songIRQAddr & 0xff00) >> 8

	if brkAfterInit:
		driver[0xa] = 0x0 # BRK

	return ''.join(chr(b) for b in driver)

class SIDFile(object):
	def __init__(self):
		self.pages = {} # Maps page address (high 4 bytes of addr)
		                # to 256 bytes of data.
		self.songhash = '' # hexdigest
		self.isPSID = True

	def load(self, filename, printSIDInfo = False):
		handle = open(filename, 'rb')

		ident = handle.read(4)
		if ident == 'RSID':
			self.isPSID = False
			raise BadSID("RSID format not yet supported")
		elif ident != 'PSID':
			raise BadSID("Wrong magic")
	
		self.isPSID = True

		headerv1 = handle.read(struct.calcsize(SIDHEADERv1))
		self.version, self.dataOffset, self.fileLoadAddr, self.initAddr, \
			self.playAddr, self.songCount, self.startSong, self.songSpeed, \
			self.songName, self.songComposer, self.songCopyright \
			= struct.unpack(SIDHEADERv1, headerv1)

		if self.version == 2:
			headerv2 = handle.read(struct.calcsize(SIDHEADERv2))
			v2Flags, self.startPage, self.endPage, v2Reserved = \
					struct.unpack(SIDHEADERv2, headerv2)
			if(v2Flags & 1):
				raise BadSID("Sidplayer MUS not supported")
			if(v2Flags & 2):
				raise BadSID("PlaySID samples not supported")
			v2PlaybackSpeed = (v2Flags & 0xc) >> 2
			if v2PlaybackSpeed == 0:
				# Don't really care, assume PAL
				self.refreshHz = 50 # PAL
			elif v2PlaybackSpeed == 1:
				self.refreshHz = 50 # PAL
			elif v2PlaybackSpeed == 2:
				self.refreshHz = 60 # NTSC
			elif v2PlaybackSpeed == 3:
				raise BadSID("Dual PAL and NTSC songs not supported")
			v2SIDHardware = (v2Flags & 0x30) >> 4
			self.hardwareVersion = {0: 'Unknown', 1: '6581', 2: '6850', 3: '6581'} \
					[v2SIDHardware]

		elif self.version != 1:
			raise BadSID("Unknown SID version (%d)" % (self.version))
			self.refreshHz = 50 # FIXME
			self.hardwareVersion = '6581'

		if self.fileLoadAddr == 0:
			# Load addr is first two bytes of SID
			self.loadAddr = struct.unpack("<H", handle.read(2))[0]
		else:
			self.loadAddr = self.fileLoadAddr

		if printSIDInfo:
			print "SID name: %s" % (self.songName)
			print "Composer: %s" % (self.songComposer)
			print "(c)     : %s" % (self.songCopyright)
			print "Version : %d" % (self.version)
			print "Addrs   : %04x(%04x) %04x %04x" % (self.loadAddr, self.fileLoadAddr,
					self.initAddr, self.playAddr)
			print "Songs   : %d" % (self.songCount)
			print "Speed   : %x" % (self.songSpeed)
			print "Hz      : %d" % (self.refreshHz)
			print 'Hardware: %s' % (self.hardwareVersion)

		# Read the file in 256-byte chunks.
		if self.loadAddr & 0xFF != 0:
			# We can only upload in page chunks, so construct a fake page.
			pageBytes = 256 - (self.loadAddr & 0xff)
			data = '\0' * (256 - pageBytes)
			data += handle.read(pageBytes)
			address = self.loadAddr & 0xff00
		else:
			address = self.loadAddr
			data = handle.read(256)

		hasher = hashlib.sha1()
		while data:
			assert address & 0xff == 0
			hasher.update(data)
			self.pages[address] = data

			if len(data) < 256:
				break

			address += 256
			data = handle.read(256)

		handle.close()
		self.songhash = hasher.hexdigest()

